import json
from pathlib import Path
from typing import Tuple

import click

from ape.cli import ape_cli_context


@click.group()
def cli():
    """
    Package management tools
    """


def _echo_no_packages(project: bool):
    message = "No packages installed"
    if project:
        message = f"{message} for this project"

    click.echo(f"{message}.")


@cli.command("list")
@ape_cli_context()
@click.option(
    "_all", "--all", is_flag=True, help="Include packages not referenced by the local project"
)
def _list(cli_ctx, _all):
    """
    List installed packages
    """

    packages = []
    packages_folder = cli_ctx.dependency_manager.DATA_FOLDER / "packages"
    if _all:
        if not packages_folder.is_dir():
            _echo_no_packages(False)
            return

        for dependency in packages_folder.iterdir():
            base_item = {"name": dependency.name}
            for version_dir in dependency.iterdir():
                item = {
                    "version": version_dir.name,
                    **base_item,
                    "compiled": _check_compiled(version_dir),
                }
                packages.append(item)

    else:
        # Limit to local project.
        for dependency in cli_ctx.config_manager.dependencies:
            item = {"name": dependency.name, "version": dependency.version_id, "compiled": False}

            # Check if compiled.
            if packages_folder.is_dir():
                for package_dir in packages_folder.iterdir():
                    if package_dir.is_dir() and package_dir.name == dependency.name:
                        for version_dir in package_dir.iterdir():
                            versions = [dependency.version_id]
                            if versions[0].startswith("v"):
                                versions.append(dependency.version_id[1:])
                            else:
                                versions.append(f"v{dependency.version_id}")

                            if version_dir.is_dir() and version_dir.name in versions:
                                item["compiled"] = _check_compiled(version_dir)

            packages.append(item)

    if not packages:
        _echo_no_packages(not _all)
        return

    # Output gathered packages.
    longest_name = max([4, *[len(p["name"]) for p in packages]])
    longest_version = max([7, *[len(p["version"]) for p in packages]])
    tab = "  "

    header_name_space = ((longest_name - len("NAME")) + 2) * " "
    version_name_space = ((longest_version - len("VERSION")) + 2) * " "

    def get_package_str(_package) -> str:
        name = click.style(_package["name"], bold=True)
        version = _package["version"]
        compiled = (
            click.style(_package["compiled"], fg="green") if _package.get("compiled") else "-"
        )
        spacing_name = ((longest_name - len(_package["name"])) + len(tab)) * " "
        spacing_version = ((longest_version - len(version)) + len(tab)) * " "
        return f"{name}{spacing_name}{version}{spacing_version + compiled}"

    def rows():
        yield f"NAME{header_name_space}VERSION{version_name_space}COMPILED\n"
        for _package in packages:
            yield f"{get_package_str(_package)}\n"

    if len(packages) > 16:
        click.echo_via_pager(rows())
    else:
        for row in rows():
            click.echo(row.strip())


def _check_compiled(version_dir: Path) -> bool:
    file = next(version_dir.iterdir(), None)
    return (
        bool(json.loads(file.read_text()).get("contractTypes"))
        if file and file.is_file()
        else False
    )


def _package_callback(ctx, param, value):
    if value is None:
        # Install all packages from local project.
        return None

    elif value.startswith("gh:"):
        # Is a GitHub style dependency
        return {"github": value[3:]}

    elif value.startswith("npm:"):
        # Is an NPM style dependency
        return {"npm:": value[4:]}

    elif value == ".":
        return value

    # Check if is a local package.
    try:
        path = Path(value).absolute()
    except Exception:
        path = None

    if path is not None and path.exists():
        # Is a local package somewhere.
        if path.is_file() and path.name == "ape-config.yaml":
            path = path.parent

        return {"local": path.as_posix()}

    elif ":" in value:
        # Catch-all for unknown dependency types that may exist.
        parts = value.split(":")
        return {parts[0]: parts[1]}

    raise click.BadArgumentUsage(f"Unknown package '{value}'.")


@cli.command()
@ape_cli_context()
@click.argument("package", nargs=1, required=False, callback=_package_callback)
@click.option("--name", help="The name of the dependency", metavar="NAME")
@click.option("--version", help="The dependency's version", metavar="VERSION")
@click.option(
    "--ref",
    help="A reference flag, used for GitHub branches or tags instead of version",
    metavar="REF",
)
@click.option("--force", "-f", help="Force a re-install", is_flag=True)
def install(cli_ctx, package, name, version, ref, force):
    """
    Download and cache packages
    """

    log_name = None

    if not package or package == ".":
        # `ape pm install`: Load all dependencies from current package.
        cli_ctx.project_manager.load_dependencies(use_cache=not force)

    elif name:
        # `ape pm install <package>`: Is a specific package.
        data = {"name": name, **package}
        if version is not None:
            data["version"] = version
        if ref is not None:
            data["ref"] = ref

        dependency_obj = cli_ctx.dependency_manager.decode_dependency(data)
        dependency_obj.extract_manifest(use_cache=not force)
        log_name = f"{dependency_obj.name}@{dependency_obj.version_id}"

    else:
        # This is **not** the local project, but no --name was given.
        # NOTE: `--version` is not required when using local dependencies.
        cli_ctx.abort("Must provide --name")

    if log_name:
        cli_ctx.logger.success(f"Package '{log_name}' installed.")
    else:
        cli_ctx.logger.success("All project packages installed.")


@cli.command()
@ape_cli_context()
@click.argument("package", nargs=1, required=True)
@click.argument("versions", nargs=-1, required=False)
@click.option(
    "-y", "--yes", is_flag=True, help="Automatically confirm the removal of the package(s)"
)
def remove(cli_ctx, package, versions, yes):
    """
    Remove a package

    This command removes a package from the installed packages.

    If specific versions are provided, only those versions of the package will be
    removed. If no versions are provided, the command will prompt you to choose
    versions to remove. You can also choose to remove all versions of the package.

    Examples:\n
    - Remove specific versions: ape pm remove <PackageName> "1.0.0" "2.0.0"\n
    - Prompt to choose versions: ape pm remove <PackageName>\n
    - Remove all versions: ape pm remove <PackageName> -y
    """
    package_dir = cli_ctx.dependency_manager.DATA_FOLDER / "packages" / package
    if not package_dir.is_dir():
        cli_ctx.abort(f"Package '{package}' is not installed.")

    # Remove multiple versions if no version is specified
    versions_to_remove = versions if versions else []
    if len(versions_to_remove) == 1 and versions_to_remove[0] == "all":
        versions_to_remove = [d.name for d in package_dir.iterdir() if d.is_dir()]

    elif not versions_to_remove:
        available_versions = [d.name for d in package_dir.iterdir() if d.is_dir()]
        if not available_versions:
            cli_ctx.abort(f"No installed versions of package '{package}' found.")

        # If there is only one version, use that.
        if len(available_versions) == 1 or yes:
            versions_to_remove = available_versions

        else:
            version_prompt = (
                f"Which versions of package '{package}' do you want to remove? "
                f"{available_versions} (separate multiple versions with comma, or 'all')"
            )
            versions_input = click.prompt(version_prompt)
            if versions_input.strip() == "all":
                versions_to_remove = available_versions
            else:
                versions_to_remove = [v.strip() for v in versions_input.split(",") if v.strip()]

            # Prevents a double-prompt.
            yes = True

    if not versions_to_remove:
        cli_ctx.logger.info("No versions selected for removal.")
        return

    # Remove all the versions specified
    for version in versions_to_remove:
        if not (package_dir / version).is_dir() and not (package_dir / f"v{version}").is_dir():
            cli_ctx.logger.warning(
                f"Version '{version}' of package '{package_dir.name}' is not installed."
            )
            continue

        elif yes or click.confirm(
            f"Are you sure you want to remove version '{version}' of package '{package}'?"
        ):
            cli_ctx.project_manager.remove_dependency(package_dir.name, versions=[version])
            cli_ctx.logger.success(f"Version '{version}' of package '{package_dir.name}' removed.")


@cli.command()
@ape_cli_context()
@click.argument("name", nargs=1, required=False)
@click.option("--version", help="The dependency version", metavar="VERSION")
@click.option("--force", "-f", help="Force a re-compile", is_flag=True)
def compile(cli_ctx, name, version, force):
    """
    Compile a package
    """

    if not name:
        # Compile all local project dependencies.
        for dep_name, versions in cli_ctx.project_manager.dependencies.items():
            for version, dependency in versions.items():
                log_line = dep_name
                if version != "local":
                    log_line += f"@{version}"

                try:
                    dependency.compile(use_cache=not force)
                except Exception as err:
                    cli_ctx.logger.error(err)
                else:
                    cli_ctx.logger.success(f"Package '{log_line}' compiled.")

        return

    elif name not in cli_ctx.project_manager.dependencies:
        cli_ctx.abort(f"Dependency '{name}' unknown. Is it installed?")

    if not (versions := cli_ctx.project_manager.dependencies[name]):
        # This shouldn't happen.
        cli_ctx.abort("No versions.")

    if not version and len(versions) == 1:
        # Version is not specified but we can use the only existing version.
        version = list(versions.keys())[0]

    elif not version:
        cli_ctx.abort("Please specify --version.")

    version_opts: Tuple
    if version == "local":
        version_opts = (version,)
    elif version.startswith("v"):
        version_opts = (f"v{version}", str(version))
    else:
        version_opts = (str(version), str(version[1:]))

    version_found = None
    for version_i in version_opts:
        if version_i in versions:
            version_found = version_i
            break

    if not version_found:
        cli_ctx.abort(f"Version '{version}' for dependency '{name}' not found. Is it installed?")

    dependency = versions[version_found]

    try:
        dependency.compile(use_cache=not force)
    except Exception as err:
        cli_ctx.logger.error(err)
    else:
        log_line = name
        if version_found and version_found != "local":
            log_line = f"{log_line}@{version_found}"

        cli_ctx.logger.success(f"Package '{log_line}' compiled.")



import threading
import time
from distutils.dir_util import copy_tree
from pathlib import Path
from typing import Optional, cast

import pytest
from ethpm_types import ContractType, HexBytes, MethodABI

import ape
from ape.api import TransactionAPI
from ape.contracts import ContractContainer, ContractInstance
from ape.contracts.base import ContractCallHandler
from ape.exceptions import ChainError, ContractLogicError
from ape.logging import LogLevel
from ape.logging import logger as _logger
from ape.types import AddressType, ContractLog
from ape_ethereum.proxies import minimal_proxy as _minimal_proxy_container

PROJECT_PATH = Path(__file__).parent
CONTRACTS_FOLDER = PROJECT_PATH / "data" / "contracts" / "ethereum" / "local"


@pytest.fixture(scope="session")
def get_contract_type():
    def fn(name: str) -> ContractType:
        return ContractType.parse_file(CONTRACTS_FOLDER / f"{name}.json")

    return fn


ALIAS_2 = "__FUNCTIONAL_TESTS_ALIAS_2__"
TEST_ADDRESS = cast(AddressType, "0xd8dA6BF26964aF9D7eEd9e03E53415D37aA96045")
BASE_PROJECTS_DIRECTORY = (Path(__file__).parent / "data" / "projects").absolute()
PROJECT_WITH_LONG_CONTRACTS_FOLDER = BASE_PROJECTS_DIRECTORY / "LongContractsFolder"
APE_PROJECT_FOLDER = BASE_PROJECTS_DIRECTORY / "ApeProject"
BASE_SOURCES_DIRECTORY = (Path(__file__).parent / "data/sources").absolute()

CALL_WITH_STRUCT_INPUT = MethodABI.parse_obj(
    {
        "type": "function",
        "name": "getTradeableOrderWithSignature",
        "stateMutability": "view",
        "inputs": [
            {"name": "owner", "type": "address", "internalType": "address"},
            {
                "name": "params",
                "type": "tuple",
                "components": [
                    {
                        "name": "handler",
                        "type": "address",
                        "internalType": "contract IConditionalOrder",
                    },
                    {"name": "salt", "type": "bytes32", "internalType": "bytes32"},
                    {"name": "staticInput", "type": "bytes", "internalType": "bytes"},
                ],
                "internalType": "struct IConditionalOrder.ConditionalOrderParams",
            },
            {"name": "offchainInput", "type": "bytes", "internalType": "bytes"},
            {"name": "proof", "type": "bytes32[]", "internalType": "bytes32[]"},
        ],
        "outputs": [
            {
                "name": "order",
                "type": "tuple",
                "components": [
                    {"name": "sellToken", "type": "address", "internalType": "contract IERC20"},
                    {"name": "buyToken", "type": "address", "internalType": "contract IERC20"},
                    {"name": "receiver", "type": "address", "internalType": "address"},
                    {"name": "sellAmount", "type": "uint256", "internalType": "uint256"},
                    {"name": "buyAmount", "type": "uint256", "internalType": "uint256"},
                    {"name": "validTo", "type": "uint32", "internalType": "uint32"},
                    {"name": "appData", "type": "bytes32", "internalType": "bytes32"},
                    {"name": "feeAmount", "type": "uint256", "internalType": "uint256"},
                    {"name": "kind", "type": "bytes32", "internalType": "bytes32"},
                    {"name": "partiallyFillable", "type": "bool", "internalType": "bool"},
                    {"name": "sellTokenBalance", "type": "bytes32", "internalType": "bytes32"},
                    {"name": "buyTokenBalance", "type": "bytes32", "internalType": "bytes32"},
                ],
                "internalType": "struct GPv2Order.Data",
            },
            {"name": "signature", "type": "bytes", "internalType": "bytes"},
        ],
    }
)
METHOD_WITH_STRUCT_INPUT = MethodABI.parse_obj(
    {
        "type": "function",
        "name": "getTradeableOrderWithSignature",
        "stateMutability": "view",
        "inputs": [
            {"name": "owner", "type": "address", "internalType": "address"},
            {
                "name": "params",
                "type": "tuple",
                "components": [
                    {
                        "name": "handler",
                        "type": "address",
                        "internalType": "contract IConditionalOrder",
                    },
                    {"name": "salt", "type": "bytes32", "internalType": "bytes32"},
                    {"name": "staticInput", "type": "bytes", "internalType": "bytes"},
                ],
                "internalType": "struct IConditionalOrder.ConditionalOrderParams",
            },
            {"name": "offchainInput", "type": "bytes", "internalType": "bytes"},
            {"name": "proof", "type": "bytes32[]", "internalType": "bytes32[]"},
        ],
        "outputs": [
            {
                "name": "order",
                "type": "tuple",
                "components": [
                    {"name": "sellToken", "type": "address", "internalType": "contract IERC20"},
                    {"name": "buyToken", "type": "address", "internalType": "contract IERC20"},
                    {"name": "receiver", "type": "address", "internalType": "address"},
                    {"name": "sellAmount", "type": "uint256", "internalType": "uint256"},
                    {"name": "buyAmount", "type": "uint256", "internalType": "uint256"},
                    {"name": "validTo", "type": "uint32", "internalType": "uint32"},
                    {"name": "appData", "type": "bytes32", "internalType": "bytes32"},
                    {"name": "feeAmount", "type": "uint256", "internalType": "uint256"},
                    {"name": "kind", "type": "bytes32", "internalType": "bytes32"},
                    {"name": "partiallyFillable", "type": "bool", "internalType": "bool"},
                    {"name": "sellTokenBalance", "type": "bytes32", "internalType": "bytes32"},
                    {"name": "buyTokenBalance", "type": "bytes32", "internalType": "bytes32"},
                ],
                "internalType": "struct GPv2Order.Data",
            },
            {"name": "signature", "type": "bytes", "internalType": "bytes"},
        ],
    }
)


class _ContractLogicError(ContractLogicError):
    pass


@pytest.hookimpl(trylast=True, hookwrapper=True)
def pytest_collection_finish(session):
    with ape.networks.parse_network_choice("::test"):
        # Sets the active provider
        yield


@pytest.fixture
def mock_web3(mocker):
    return mocker.MagicMock()


@pytest.fixture
def mock_transaction(mocker):
    return mocker.MagicMock(spec=TransactionAPI)


@pytest.fixture(scope="session")
def project_path():
    return PROJECT_PATH


@pytest.fixture(scope="session")
def contracts_folder():
    return CONTRACTS_FOLDER


@pytest.fixture(scope="session")
def address():
    return TEST_ADDRESS


@pytest.fixture
def second_keyfile_account(sender, keyparams, temp_accounts_path, temp_keyfile_account_ctx):
    with temp_keyfile_account_ctx(temp_accounts_path, ALIAS_2, keyparams, sender) as account:
        yield account


@pytest.fixture(scope="session")
def solidity_contract_type(get_contract_type) -> ContractType:
    return get_contract_type("solidity_contract")


@pytest.fixture(scope="session")
def solidity_contract_container(solidity_contract_type) -> ContractContainer:
    return ContractContainer(contract_type=solidity_contract_type)


@pytest.fixture
def solidity_contract_instance(
    owner, solidity_contract_container, networks_connected_to_tester
) -> ContractInstance:
    return owner.deploy(solidity_contract_container, 0)


@pytest.fixture(scope="session")
def vyper_contract_type(get_contract_type) -> ContractType:
    return get_contract_type("vyper_contract")


@pytest.fixture(scope="session")
def solidity_fallback_contract_type(get_contract_type) -> ContractType:
    return get_contract_type("sol_fallback_and_receive")


@pytest.fixture(scope="session")
def vyper_fallback_contract_type(get_contract_type) -> ContractType:
    return get_contract_type("vy_default")


@pytest.fixture(scope="session")
def vyper_contract_container(vyper_contract_type) -> ContractContainer:
    return ContractContainer(contract_type=vyper_contract_type)


@pytest.fixture(scope="session")
def solidity_fallback_container(solidity_fallback_contract_type) -> ContractContainer:
    return ContractContainer(contract_type=solidity_fallback_contract_type)


@pytest.fixture(scope="session")
def vyper_fallback_container(vyper_fallback_contract_type) -> ContractContainer:
    return ContractContainer(contract_type=vyper_fallback_contract_type)


@pytest.fixture
def vyper_contract_instance(
    owner, vyper_contract_container, networks_connected_to_tester
) -> ContractInstance:
    return owner.deploy(vyper_contract_container, 0, required_confirmations=0)


@pytest.fixture
def solidity_fallback_contract(owner, solidity_fallback_container):
    return owner.deploy(solidity_fallback_container)


@pytest.fixture
def vyper_fallback_contract(owner, vyper_fallback_container):
    return owner.deploy(vyper_fallback_container)


@pytest.fixture(scope="session")
def reverts_contract_type(get_contract_type) -> ContractType:
    return get_contract_type("reverts_contract")


@pytest.fixture(scope="session")
def sub_reverts_contract_type(get_contract_type) -> ContractType:
    return get_contract_type("sub_reverts")


@pytest.fixture(scope="session")
def reverts_contract_container(reverts_contract_type) -> ContractContainer:
    return ContractContainer(contract_type=reverts_contract_type)


@pytest.fixture(scope="session")
def sub_reverts_contract_container(sub_reverts_contract_type) -> ContractContainer:
    return ContractContainer(contract_type=sub_reverts_contract_type)


@pytest.fixture
def reverts_contract_instance(
    owner, reverts_contract_container, sub_reverts_contract_instance, eth_tester_provider
) -> ContractInstance:
    return owner.deploy(
        reverts_contract_container, sub_reverts_contract_instance, required_confirmations=0
    )


@pytest.fixture(params=("solidity", "vyper"))
def contract_container(
    request, solidity_contract_container, vyper_contract_container, networks_connected_to_tester
):
    return solidity_contract_container if request.param == "solidity" else vyper_contract_container


@pytest.fixture(params=("solidity", "vyper"))
def contract_instance(
    eth_tester_provider, request, solidity_contract_instance, vyper_contract_instance
):
    return solidity_contract_instance if request.param == "solidity" else vyper_contract_instance


@pytest.fixture(params=("solidity", "vyper"))
def fallback_contract(
    eth_tester_provider, request, solidity_fallback_contract, vyper_fallback_contract
):
    return solidity_fallback_contract if request.param == "solidity" else vyper_fallback_contract


@pytest.fixture
def ds_note_test_contract(eth_tester_provider, vyper_contract_type, owner, get_contract_type):
    contract_type = get_contract_type("ds_note_test")
    contract_container = ContractContainer(contract_type=contract_type)
    return contract_container.deploy(sender=owner)


@pytest.fixture
def project_with_contract(temp_config):
    with temp_config() as project:
        copy_tree(str(APE_PROJECT_FOLDER), str(project.path))
        yield project


@pytest.fixture
def project_with_source_files_contract(temp_config):
    bases_source_dir = BASE_SOURCES_DIRECTORY
    project_source_dir = APE_PROJECT_FOLDER

    with temp_config() as project:
        copy_tree(str(project_source_dir), str(project.path))
        copy_tree(str(bases_source_dir), f"{project.path}/contracts/")
        yield project


@pytest.fixture
def clean_contracts_cache(chain):
    original_cached_contracts = chain.contracts._local_contract_types
    chain.contracts._local_contract_types = {}
    yield
    chain.contracts._local_contract_types = original_cached_contracts


@pytest.fixture
def project_with_dependency_config(temp_config):
    dependencies_config = {
        "dependencies": [
            {
                "local": str(PROJECT_WITH_LONG_CONTRACTS_FOLDER),
                "name": "testdependency",
                "contracts_folder": "source/v0.1",
            }
        ]
    }
    with temp_config(dependencies_config) as project:
        yield project


@pytest.fixture(scope="session")
def base_projects_directory():
    return BASE_PROJECTS_DIRECTORY


@pytest.fixture(scope="session")
def mainnet_contract(chain):
    def contract_getter(address):
        path = (
            Path(__file__).parent
            / "data"
            / "contracts"
            / "ethereum"
            / "mainnet"
            / f"{address}.json"
        )
        contract = ContractType.parse_file(path)
        chain.contracts._local_contract_types[address] = contract
        return contract

    return contract_getter


@pytest.fixture(scope="session")
def ds_note():
    return {
        "address": "0x35D1b3F3D7966A1DFe207aa4514C12a259A0492B",
        "topics": [
            HexBytes("0x7608870300000000000000000000000000000000000000000000000000000000"),
            HexBytes("0x5946492d41000000000000000000000000000000000000000000000000000000"),
            HexBytes("0x0000000000000000000000000abb839063ef747c8432b2acc60bf8f70ec09a45"),
            HexBytes("0x0000000000000000000000000abb839063ef747c8432b2acc60bf8f70ec09a45"),
        ],
        "data": "0x000000000000000000000000000000000000000000000000000000000000002000000000000000000000000000000000000000000000000000000000000000e0760887035946492d410000000000000000000000000000000000000000000000000000000000000000000000000000000abb839063ef747c8432b2acc60bf8f70ec09a450000000000000000000000000abb839063ef747c8432b2acc60bf8f70ec09a450000000000000000000000000abb839063ef747c8432b2acc60bf8f70ec09a450000000000000000000000000000000000000000000000000000000000000000fffffffffffffffffffffffffffffffffffffffffffa050e82a57b7fc6b6020c00000000000000000000000000000000000000000000000000000000",  # noqa: E501
        "blockNumber": 14623434,
        "transactionHash": HexBytes(
            "0xa322a9fd0e627e22bfe1b0877cca1d1f2e697d076007231d0b7a366d1a0fdd51"
        ),
        "transactionIndex": 333,
        "blockHash": HexBytes("0x0fd77b0af3fa471aa040a02d4fcd1ec0a35122a4166d0bb7c31354e23823de49"),
        "logIndex": 376,
        "removed": False,
    }


@pytest.fixture
def chain_that_mined_5(chain):
    chain.mine(5)
    return chain


class PollDaemonThread(threading.Thread):
    def __init__(self, name, poller, handler, stop_condition, *args, **kwargs):
        kwargs_dict = dict(**kwargs)
        kwargs_dict["name"] = f"ape_poll_{name}"
        super().__init__(*args, **kwargs_dict)
        self._poller = poller
        self._handler = handler
        self._do_stop = stop_condition
        self._exception = None
        self._max_iterations = 100

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.stop()

    def run(self):
        try:
            self._run_until_stop(timeout_iterations=self._max_iterations)
        except Exception as err:
            self._exception = err

    def stop(self):
        time.sleep(1)
        self.join()

        # Attempt to wait for stop condition
        if not self._do_stop():
            self._run_until_stop(timeout_iterations=10)

    def join(self, timeout=None):
        super().join(timeout=timeout)
        if self._exception and not self._do_stop():
            # Only raise if error-ed before hitting stop condition
            raise self._exception

    def _run_until_stop(self, timeout_iterations: Optional[int] = None):
        iterations = 0
        while True:
            if self._do_stop():
                return

            try:
                self._handler(next(self._poller))
            except ChainError:
                # Check if can stop once more before exiting
                if self._do_stop():
                    return

                raise  # The timeout ChainError

            time.sleep(1)
            if timeout_iterations is None:
                continue

            elif iterations >= timeout_iterations:
                return

            iterations += 1


@pytest.fixture
def PollDaemon():
    return PollDaemonThread


@pytest.fixture
def assert_log_values(contract_instance):
    def _assert_log_values(
        log: ContractLog,
        number: int,
        previous_number: Optional[int] = None,
        address: Optional[AddressType] = None,
    ):
        assert log.contract_address == address or contract_instance.address
        assert isinstance(log.b, bytes)
        expected_previous_number = number - 1 if previous_number is None else previous_number
        assert log.prevNum == expected_previous_number, "Event param 'prevNum' has unexpected value"
        assert log.newNum == number, "Event param 'newNum' has unexpected value"
        assert log.dynData == "Dynamic"
        assert log.dynIndexed == HexBytes(
            "0x9f3d45ac20ccf04b45028b8080bb191eab93e29f7898ed43acf480dd80bba94d"
        )

    return _assert_log_values


@pytest.fixture
def remove_disk_writes_deployments(chain):
    if chain.contracts._deployments_mapping_cache.is_file():
        chain.contracts._deployments_mapping_cache.unlink()

    yield

    if chain.contracts._deployments_mapping_cache.is_file():
        chain.contracts._deployments_mapping_cache.unlink()


@pytest.fixture(scope="session")
def logger():
    return _logger


@pytest.fixture
def use_debug(logger):
    initial_level = logger.level
    logger.set_level(LogLevel.DEBUG)
    yield
    logger.set_level(initial_level)


@pytest.fixture
def dummy_live_network(chain):
    original_network = chain.provider.network.name
    chain.provider.network.name = "goerli"
    yield chain.provider.network
    chain.provider.network.name = original_network


@pytest.fixture(scope="session")
def proxy_contract_container(get_contract_type):
    return ContractContainer(get_contract_type("proxy"))


@pytest.fixture(scope="session")
def calldata():
    return HexBytes(
        "0x3fb5c1cb00000000000000000000000000000000000000000000000000000000000000de"
    )  # setNumber(222)


@pytest.fixture(scope="session")
def calldata_with_address():
    return HexBytes(
        "0x9e6b154b00000000000000000000000000000000000000000000000000000000000000de"
        "000000000000000000000000f7f78379391c5df2db5b66616d18ff92edb82022"
    )  # setNumber(222, 0xf7f78379391c5df2db5b66616d18ff92edb82022)


@pytest.fixture(scope="session")
def unique_calldata():
    return HexBytes(
        "0xacab48d8"
        "0000000000000000000000000000000000000000000000000000000000000000"
        "0000000000000000000000000000000000000000000000000000000000000001"
        "0000000000000000000000000000000000000000000000000000000000000002"
        "0000000000000000000000000000000000000000000000000000000000000003"
        "0000000000000000000000000000000000000000000000000000000000000004"
        "0000000000000000000000000000000000000000000000000000000000000005"
        "0000000000000000000000000000000000000000000000000000000000000006"
        "0000000000000000000000000000000000000000000000000000000000000007"
        "0000000000000000000000000000000000000000000000000000000000000008"
        "0000000000000000000000000000000000000000000000000000000000000009"
    )  # functionWithUniqueArguments(0, 1, 2, 3, 4, 5, 6, 7, 8, 9)


@pytest.fixture
def leaf_contract(eth_tester_provider, owner, get_contract_type):
    ct = get_contract_type("contract_c")
    return owner.deploy(ContractContainer(ct))


@pytest.fixture
def middle_contract(eth_tester_provider, owner, get_contract_type, leaf_contract):
    ct = get_contract_type("contract_b")
    return owner.deploy(ContractContainer(ct), leaf_contract)


@pytest.fixture
def contract_with_call_depth(
    owner, eth_tester_provider, get_contract_type, leaf_contract, middle_contract
):
    contract = ContractContainer(get_contract_type("contract_a"))
    return owner.deploy(contract, middle_contract, leaf_contract)


@pytest.fixture
def sub_reverts_contract_instance(owner, sub_reverts_contract_container, eth_tester_provider):
    return owner.deploy(sub_reverts_contract_container, required_confirmations=0)


@pytest.fixture(scope="session")
def error_contract_container(get_contract_type):
    ct = get_contract_type("has_error")
    return ContractContainer(ct)


@pytest.fixture
def error_contract(owner, error_contract_container, eth_tester_provider):
    _ = eth_tester_provider  # Ensure uses eth tester
    return owner.deploy(error_contract_container, 1)


@pytest.fixture
def vyper_factory(owner, get_contract_type):
    return owner.deploy(ContractContainer(get_contract_type("vyper_factory")))


@pytest.fixture
def vyper_blueprint(owner, vyper_contract_container):
    receipt = owner.declare(vyper_contract_container)
    return receipt.contract_address


@pytest.fixture
def minimal_proxy_container():
    return _minimal_proxy_container


@pytest.fixture
def minimal_proxy(owner, minimal_proxy_container):
    return owner.deploy(minimal_proxy_container)


@pytest.fixture
def mock_explorer(mocker):
    explorer = mocker.MagicMock()
    explorer.name = "mock"  # Needed for network data serialization.
    return explorer


@pytest.fixture
def call_abi_with_struct_input():
    return CALL_WITH_STRUCT_INPUT


@pytest.fixture
def fake_contract(mocker):
    # Only needed for initialization; never used.
    return mocker.MagicMock()


@pytest.fixture
def call_handler_with_struct_input(fake_contract, call_abi_with_struct_input):
    abi = call_abi_with_struct_input
    return ContractCallHandler(contract=fake_contract, abis=[abi])


@pytest.fixture(scope="session")
def struct_input_for_call(owner):
    return [owner, [owner, b"skip", b"skip"], b"skip", [b"skip"]]


@pytest.fixture(scope="session")
def output_from_struct_input_call():
    # Expected when using `struct_input_for_call`.
    return HexBytes(
        "0x26e0a1960000000000000000000000001e59ce931b4cfea3fe4b875411e280e173cb7a9c"
        "00000000000000000000000000000000000000000000000000000000000000800000000000"
        "00000000000000000000000000000000000000000000000000012000000000000000000000"
        "000000000000000000000000000000000000000001600000000000000000000000001e59ce"
        "931b4cfea3fe4b875411e280e173cb7a9c736b697000000000000000000000000000000000"
        "00000000000000000000000000000000000000000000000000000000000000000000000000"
        "00000000000060000000000000000000000000000000000000000000000000000000000000"
        "0004736b697000000000000000000000000000000000000000000000000000000000000000"
        "0000000000000000000000000000000000000000000000000000000004736b697000000000"
        "00000000000000000000000000000000000000000000000000000000000000000000000000"
        "00000000000000000000000000000000000001736b69700000000000000000000000000000"
        "0000000000000000000000000000"
    )


@pytest.fixture
def method_abi_with_struct_input():
    return METHOD_WITH_STRUCT_INPUT


@pytest.fixture
def mock_compiler(mocker):
    mock = mocker.MagicMock()
    mock.name = "mock"
    mock.ext = ".__mock__"

    def mock_compile(paths, base_path=None):
        mock.tracked_settings.append(mock.compiler_settings)
        result = []
        for path in paths:
            if path.suffix == mock.ext:
                name = path.stem
                code = HexBytes(123).hex()
                contract_type = ContractType(contractName=name, abi=[], deploymentBytecode=code)
                result.append(contract_type)

        return result

    mock.compile.side_effect = mock_compile
    return mock



from typing import Iterator, List, Optional

from eth_account import Account as EthAccount
from eth_account.messages import SignableMessage
from eth_utils import to_bytes

from ape.api import TestAccountAPI, TestAccountContainerAPI, TransactionAPI
from ape.types import AddressType, MessageSignature, TransactionSignature
from ape.utils import GeneratedDevAccount, generate_dev_accounts


class TestAccountContainer(TestAccountContainerAPI):
    _num_generated: int
    _accounts: List["TestAccount"]
    _mnemonic: str
    _num_of_accounts: int
    _hd_path: str

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.init()

    def init(self):
        self._num_generated = 0
        self._accounts = []
        self._mnemonic = self.config["mnemonic"]
        self._num_of_accounts = self.config["number_of_accounts"]
        self._hd_path = self.config["hd_path"]
        for index, account in enumerate(self._dev_accounts):
            self._accounts.append(
                TestAccount(
                    index=index, address_str=account.address, private_key=account.private_key
                )
            )

    def __len__(self) -> int:
        return self._num_of_accounts

    @property
    def config(self):
        return self.config_manager.get_config("test")

    @property
    def _dev_accounts(self) -> List[GeneratedDevAccount]:
        return generate_dev_accounts(
            self._mnemonic,
            number_of_accounts=self._num_of_accounts,
            hd_path_format=self._hd_path,
        )

    @property
    def aliases(self) -> Iterator[str]:
        for index in range(self._num_of_accounts):
            yield f"TEST::{index}"

    def _is_config_changed(self):
        current_mnemonic = self.config["mnemonic"]
        current_number = self.config["number_of_accounts"]
        current_hd_path = self.config["hd_path"]
        return (
            self._mnemonic != current_mnemonic
            or self._num_of_accounts != current_number
            or self._hd_path != current_hd_path
        )

    @property
    def accounts(self) -> Iterator["TestAccount"]:
        # As TestAccountManager only uses accounts property this works!
        config_changed = self._is_config_changed()
        if config_changed:
            self.init()
        for account in self._accounts:
            yield account

    def generate_account(self) -> "TestAccountAPI":
        new_index = self._num_of_accounts + self._num_generated
        self._num_generated += 1
        generated_account = generate_dev_accounts(
            self._mnemonic, 1, hd_path_format=self._hd_path, start_index=new_index
        )[0]
        acc = TestAccount(
            index=new_index,
            address_str=generated_account.address,
            private_key=generated_account.private_key,
        )
        return acc


class TestAccount(TestAccountAPI):
    index: int
    address_str: str
    private_key: str

    __test__ = False

    @property
    def alias(self) -> str:
        return f"TEST::{self.index}"

    @property
    def address(self) -> AddressType:
        return self.network_manager.ethereum.decode_address(self.address_str)

    def sign_message(self, msg: SignableMessage) -> Optional[MessageSignature]:
        signed_msg = EthAccount.sign_message(msg, self.private_key)
        return MessageSignature(
            v=signed_msg.v,
            r=to_bytes(signed_msg.r),
            s=to_bytes(signed_msg.s),
        )

    def sign_transaction(self, txn: TransactionAPI, **kwargs) -> Optional[TransactionAPI]:
        # Signs anything that's given to it
        signature = EthAccount.sign_transaction(txn.dict(), self.private_key)
        txn.signature = TransactionSignature(
            v=signature.v,
            r=to_bytes(signature.r),
            s=to_bytes(signature.s),
        )

        return txn



from ape.managers.accounts import TestAccountManager
from ape.managers.chain import ChainManager
from ape.managers.networks import NetworkManager
from ape.managers.project import ProjectManager

# Even though it is set on the class, it needs to be set _in the test_
# for some reason.
TestAccountManager.__test__ = False


def test_accounts(accounts):
    assert isinstance(accounts, TestAccountManager)


def test_chain(chain):
    assert isinstance(chain, ChainManager)


def test_contract(chain, Contract):
    assert Contract == chain.contracts.instance_at


def test_networks(networks):
    assert isinstance(networks, NetworkManager)


def test_project(project):
    assert isinstance(project, ProjectManager)


def test_built_in_fixtures(chain, capsys):
    _ = chain  # Include chain to show we can use with our shipped filters
    assert True



import shlex

import click
from click.testing import CliRunner
from eth_utils import is_hex
from IPython import get_ipython
from IPython.core.magic import Magics, line_magic, magics_class

import ape
from ape._cli import cli
from ape.exceptions import Abort, ApeException, handle_ape_exception
from ape.logging import logger
from ape.types import AddressType
from ape.utils import cached_property


@magics_class
class ApeConsoleMagics(Magics):
    @cached_property
    def ipython(self):
        if ipython := get_ipython():
            return ipython

        raise ValueError("Must be called from an IPython session.")

    @line_magic
    def ape(self, line: str = ""):
        """
        Run Ape CLI commands within an ``ape console`` session.

        Usage example::

            %ape accounts list

        """
        runner = CliRunner()
        if "console" in [x.strip("\"' \t\n") for x in shlex.split(line)]:
            # Prevent running console within console because usually bad
            # stuff happens when you try to do this.
            raise ValueError("Unable to run `console` within a console.")

        result = runner.invoke(cli, line)
        if result.output:
            click.echo(result.output)

        return result

    @line_magic
    def bal(self, line: str = ""):
        """
        Show an account balance in human-readable form.

        Usage example::

            account = accounts.load("me")
            %bal account
        """

        if not line:
            raise ValueError("Missing argument.")

        provider = ape.networks.provider
        ecosystem = provider.network.ecosystem
        result = eval(line, self.ipython.user_global_ns, self.ipython.user_ns)
        if isinstance(result, str) and not is_hex(result):
            # Check if is an account alias.
            address = ape.accounts.load(result).address
        else:
            address = ape.convert(result, AddressType)

        decimals = ecosystem.fee_token_decimals
        symbol = ecosystem.fee_token_symbol
        balance = provider.get_balance(address)
        return f"{round(balance / 10 ** decimals, 8)} {symbol}"


def custom_exception_handler(self, etype, value, tb, tb_offset=None):
    if not handle_ape_exception(value, [self.user_ns["project"].path]):
        logger.error(Abort.from_ape_exception(value).format_message())


def load_ipython_extension(ipython):
    ipython.register_magics(ApeConsoleMagics)
    ipython.set_custom_exc((ApeException,), custom_exception_handler)



import json
from pathlib import Path
from typing import List, Optional, Set

from eth_utils import is_0x_prefixed
from ethpm_types import ContractType, HexBytes

from ape.api import CompilerAPI
from ape.exceptions import CompilerError, ContractLogicError
from ape.logging import logger
from ape.utils import get_relative_path


class InterfaceCompiler(CompilerAPI):
    @property
    def name(self) -> str:
        return "ethpm"

    def get_versions(self, all_paths: List[Path]) -> Set[str]:
        # NOTE: This bypasses the serialization of this compiler into the package manifest's
        #       ``compilers`` field. You should not do this with a real compiler plugin.
        return set()

    def compile(
        self, filepaths: List[Path], base_path: Optional[Path] = None
    ) -> List[ContractType]:
        filepaths.sort()  # Sort to assist in reproducing consistent results.
        contract_types: List[ContractType] = []
        for path in filepaths:
            source_path = (
                get_relative_path(path, base_path) if base_path and path.is_absolute() else path
            )
            source_id = str(source_path)

            try:
                # NOTE: Always set the source ID to the source of the JSON file
                #   to avoid manifest corruptions later on.
                contract_type = self.compile_code(
                    path.read_text(),
                    base_path=base_path,
                    sourceId=source_id,
                )

                # NOTE: Try getting name/ ID from code-JSON first.
                #   That's why this is not part of `contract_type_overrides`.
                if not contract_type.name:
                    contract_type.name = path.stem

            except CompilerError:
                logger.warning(f"Unable to parse {ContractType.__name__} from '{source_id}'.")
                continue

            contract_types.append(contract_type)

        return contract_types

    def compile_code(
        self,
        code: str,
        base_path: Optional[Path] = None,
        **kwargs,
    ) -> ContractType:
        data = json.loads(code)
        if isinstance(data, list):
            # ABI JSON list
            contract_type_data = {"abi": data, **kwargs}

        elif isinstance(data, dict) and (
            "contractName" in data or "abi" in data or "sourceId" in data
        ):
            # Raw contract type JSON or raw compiler output.
            contract_type_data = {**data, **kwargs}

            if (
                "deploymentBytecode" not in contract_type_data
                or "runtimeBytecode" not in contract_type_data
            ):
                if "bin" in contract_type_data:
                    # Handle raw Solidity output.
                    deployment_bytecode = data["bin"]
                    runtime_bytecode = data["bin"]

                elif "bytecode" in contract_type_data or "bytecode_runtime" in contract_type_data:
                    # Handle raw Vyper output.
                    deployment_bytecode = contract_type_data.pop("bytecode", None)
                    runtime_bytecode = contract_type_data.pop("bytecode_runtime", None)

                else:
                    deployment_bytecode = None
                    runtime_bytecode = None

                if deployment_bytecode:
                    contract_type_data["deploymentBytecode"] = {"bytecode": deployment_bytecode}
                if runtime_bytecode:
                    contract_type_data["runtimeBytecode"] = {"bytecode": runtime_bytecode}

        else:
            raise CompilerError(f"Unable to parse {ContractType.__name__}.")

        return ContractType(**contract_type_data)

    def enrich_error(self, err: ContractLogicError) -> ContractLogicError:
        if not (address := err.address) or not is_0x_prefixed(err.revert_message):
            return err

        # Check for ErrorABI.
        bytes_message = HexBytes(err.revert_message)
        selector = bytes_message[:4]
        input_data = bytes_message[4:]

        try:
            contract = self.chain_manager.contracts.instance_at(address)
        except Exception:
            return err

        if (
            not contract
            or not self.network_manager.active_provider
            or selector not in contract.contract_type.errors
        ):
            return err

        ecosystem = self.provider.network.ecosystem
        abi = contract.contract_type.errors[selector]
        inputs = ecosystem.decode_calldata(abi, input_data)
        error_class = contract.get_error_by_signature(abi.signature)
        return error_class(
            abi,
            inputs,
            txn=err.txn,
            trace=err.trace,
            contract_address=address,
            source_traceback=err.source_traceback,
        )



import json
from os import environ
from pathlib import Path
from typing import Iterator, Optional

import click
from eth_account import Account as EthAccount
from eth_keys import keys  # type: ignore
from eth_utils import to_bytes
from ethpm_types import HexBytes

from ape.api import AccountAPI, AccountContainerAPI, TransactionAPI
from ape.exceptions import AccountsError
from ape.logging import logger
from ape.types import AddressType, MessageSignature, SignableMessage, TransactionSignature


class InvalidPasswordError(AccountsError):
    """
    Raised when password to unlock an account is incorrect.
    """

    def __init__(self):
        super().__init__("Invalid password")


class AccountContainer(AccountContainerAPI):
    @property
    def _keyfiles(self) -> Iterator[Path]:
        return self.data_folder.glob("*.json")

    @property
    def aliases(self) -> Iterator[str]:
        for p in self._keyfiles:
            yield p.stem

    @property
    def accounts(self) -> Iterator[AccountAPI]:
        for keyfile in self._keyfiles:
            yield KeyfileAccount(keyfile_path=keyfile)

    def __len__(self) -> int:
        return len([*self._keyfiles])


# NOTE: `AccountAPI` is an BaseInterfaceModel
class KeyfileAccount(AccountAPI):
    keyfile_path: Path
    locked: bool = True
    __autosign: bool = False
    __cached_key: Optional[HexBytes] = None

    def __repr__(self):
        # NOTE: Prevent errors from preventing repr from working.
        try:
            address_str = f" address={self.address} "
        except Exception:
            address_str = ""

        try:
            alias_str = f" alias={self.alias} "
        except Exception:
            alias_str = ""

        return f"<{self.__class__.__name__}{address_str}{alias_str}>"

    @property
    def alias(self) -> str:
        return self.keyfile_path.stem

    @property
    def keyfile(self) -> dict:
        return json.loads(self.keyfile_path.read_text())

    @property
    def address(self) -> AddressType:
        return self.network_manager.ethereum.decode_address(self.keyfile["address"])

    @property
    def __key(self) -> HexBytes:
        if self.__cached_key is not None:
            if not self.locked:
                logger.warning("Using cached key for %s", self.alias)
                return self.__cached_key
            self.__cached_key = None

        passphrase = self._prompt_for_passphrase(default="")
        key = self.__decrypt_keyfile(passphrase)

        if click.confirm(f"Leave '{self.alias}' unlocked?"):
            self.locked = False
            self.__cached_key = key

        return key

    @property
    def public_key(self) -> HexBytes:
        if "public_key" in self.keyfile:
            return HexBytes(bytes.fromhex(self.keyfile["public_key"]))
        key = self.__key

        # Derive the public key from the private key
        pk = keys.PrivateKey(key)
        # convert from eth_keys.datatypes.PublicKey to str to make it HexBytes
        publicKey = str(pk.public_key)

        key_file_data = self.keyfile
        key_file_data["public_key"] = publicKey[2:]

        self.keyfile_path.write_text(json.dumps(key_file_data))

        return HexBytes(bytes.fromhex(publicKey[2:]))

    def unlock(self, passphrase: Optional[str] = None):
        if not passphrase:
            # Check if environment variable is available
            env_variable = f"APE_ACCOUNTS_{self.alias}_PASSPHRASE"
            passphrase = environ.get(env_variable, None)

            if passphrase:
                # Passphrase found in environment variable
                logger.info(
                    f"Using passphrase for account '{self.alias}' from environment variable"
                )
            else:
                # Passphrase not found, prompt for it
                passphrase = self._prompt_for_passphrase(
                    f"Enter passphrase to permanently unlock '{self.alias}'"
                )
        assert passphrase is not None, "Passphrase can't be 'None'"
        # Rest of the code to unlock the account using the passphrase
        self.__cached_key = self.__decrypt_keyfile(passphrase)
        self.locked = False

    def lock(self):
        self.locked = True

    def change_password(self):
        self.locked = True  # force entering passphrase to get key
        key = self.__key

        passphrase = self._prompt_for_passphrase("Create new passphrase", confirmation_prompt=True)
        self.keyfile_path.write_text(json.dumps(EthAccount.encrypt(key, passphrase)))

    def delete(self):
        passphrase = self._prompt_for_passphrase(
            f"Enter passphrase to delete '{self.alias}'", default=""
        )
        self.__decrypt_keyfile(passphrase)
        self.keyfile_path.unlink()

    def sign_message(self, msg: SignableMessage) -> Optional[MessageSignature]:
        user_approves = self.__autosign or click.confirm(f"{msg}\n\nSign: ")
        if not user_approves:
            return None

        signed_msg = EthAccount.sign_message(msg, self.__key)
        return MessageSignature(
            v=signed_msg.v,
            r=to_bytes(signed_msg.r),
            s=to_bytes(signed_msg.s),
        )

    def sign_transaction(self, txn: TransactionAPI, **kwargs) -> Optional[TransactionAPI]:
        user_approves = self.__autosign or click.confirm(f"{txn}\n\nSign: ")
        if not user_approves:
            return None

        signature = EthAccount.sign_transaction(
            txn.dict(exclude_none=True, by_alias=True), self.__key
        )
        txn.signature = TransactionSignature(
            v=signature.v,
            r=to_bytes(signature.r),
            s=to_bytes(signature.s),
        )

        return txn

    def set_autosign(self, enabled: bool, passphrase: Optional[str] = None):
        """
        Allow this account to automatically sign messages and transactions.

        Args:
            enabled (bool): ``True`` to enable, ``False`` to disable.
            passphrase (Optional[str]): Optionally provide the passphrase.
              If not provided, you will be prompted to enter it.
        """
        if enabled:
            self.unlock(passphrase=passphrase)
            logger.warning("Danger! This account will now sign any transaction it's given.")

        self.__autosign = enabled
        if not enabled:
            # Re-lock if was turning off
            self.locked = True
            self.__cached_key = None

    def _prompt_for_passphrase(self, message: Optional[str] = None, **kwargs) -> str:
        message = message or f"Enter passphrase to unlock '{self.alias}'"
        return click.prompt(
            message,
            hide_input=True,
            **kwargs,
        )

    def __decrypt_keyfile(self, passphrase: str) -> HexBytes:
        try:
            return EthAccount.decrypt(self.keyfile, passphrase)
        except ValueError as err:
            raise InvalidPasswordError() from err



import sys

from .utils import skip_projects_except

BAD_COMMAND = "not-a-name"


@skip_projects_except("script")
def test_run_unknown_script(ape_cli, runner, project):
    result = runner.invoke(ape_cli, ["run", BAD_COMMAND])
    assert result.exit_code == 2
    assert f"No such command '{BAD_COMMAND}'." in result.output


@skip_projects_except("script")
def test_run(ape_cli, runner, project):
    result = runner.invoke(ape_cli, ["run"])
    assert result.exit_code == 0, result.output
    # By default, no commands are run
    assert "Super secret script output" not in result.output

    not_part_of_test = ("output_contract_view_methods",)
    scripts = [
        s
        for s in project.scripts_folder.glob("*.py")
        if not s.name.startswith("error") and s.stem not in not_part_of_test
    ]
    for script_file in scripts:
        result = runner.invoke(ape_cli, ["run", script_file.stem], catch_exceptions=False)
        assert (
            result.exit_code == 0
        ), f"Unexpected exit code for '{script_file.name}'\n{result.output}"

        if script_file.stem.startswith("_"):
            assert "Super secret script output" not in result.output

        else:
            assert "Super secret script output" in result.output


@skip_projects_except("script")
def test_run_with_verbosity(ape_cli, runner, project):
    result = runner.invoke(ape_cli, ["run", "click", "--verbosity", "DEBUG"])
    assert result.exit_code == 0, result.output


@skip_projects_except("script")
def test_run_subdirectories(ape_cli, runner, project):
    result = runner.invoke(ape_cli, ["run"])
    assert result.exit_code == 0, result.output
    # By default, no commands are run
    assert "Super secret script output" not in result.output
    subdirectory_scripts = [
        s
        for s in (project.scripts_folder / "subdirectory").rglob("*.py")
        if not s.name.startswith("error")
    ]
    for each in subdirectory_scripts:
        result = runner.invoke(ape_cli, ["run", "subdirectory", each.stem])
        assert result.exit_code == 0
        assert "Super secret script output" in result.output


@skip_projects_except("only-script-subdirs")
def test_run_only_subdirs(ape_cli, runner, project):
    result = runner.invoke(ape_cli, ["run"])
    assert result.exit_code == 0, result.output
    # By default, no commands are run
    assert "Super secret script output" not in result.output
    subdirectory_scripts = [
        s
        for s in (project.scripts_folder / "subdirectory").rglob("*.py")
        if not s.name.startswith("error")
    ]
    for each in subdirectory_scripts:
        result = runner.invoke(ape_cli, ["run", "subdirectory", each.stem])
        assert result.exit_code == 0, f"Unexpected exit code for '{each.name}'"
        assert "Super secret script output" in result.output


@skip_projects_except("script")
def test_run_when_script_errors(ape_cli, runner, project):
    scripts = [
        s
        for s in project.scripts_folder.glob("*.py")
        if s.name.startswith("error") and not s.name.endswith("forgot_click.py")
    ]
    for script_file in scripts:
        result = runner.invoke(ape_cli, ["run", script_file.stem])
        assert (
            result.exit_code != 0
        ), f"Unexpected exit code for '{script_file.name}'.\n{result.output}"
        assert str(result.exception) == "Expected exception"


@skip_projects_except("script")
def test_run_interactive(ape_cli, runner, project):
    scripts = [
        project.scripts_folder / f"{s}.py" for s in ("error_main", "error_cli", "error_no_def")
    ]

    # Show that the variable namespace from the script is available in the console.
    user_input = "local_variable\nexit\n"

    result = runner.invoke(ape_cli, ["run", "--interactive", scripts[0].stem], input=user_input)
    assert result.exit_code == 0, result.output

    # From script: local_variable = "test foo bar"
    assert "test foo bar" in result.output


@skip_projects_except("script")
def test_run_adhoc_provider(ape_cli, runner, project):
    result = runner.invoke(
        ape_cli, ["run", "deploy", "--network", "ethereum:mainnet:http://127.0.0.1:9545"]
    )

    # Show that it attempts to connect
    assert result.exit_code == 1, result.output
    assert "No node found on 'http://127.0.0.1:9545" in result.output


@skip_projects_except("script")
def test_run_adhoc_network(ape_cli, runner, project):
    result = runner.invoke(ape_cli, ["run", "deploy", "--network", "http://127.0.0.1:9545"])

    # Show that it attempts to connect
    assert result.exit_code == 1, result.output
    assert "No node found on 'http://127.0.0.1:9545" in result.output


@skip_projects_except("script")
def test_try_run_script_missing_cli_decorator(ape_cli, runner, project):
    """
    Shows that we cannot run a script defining a `cli()` method without
    it being a click command. The script is not recognized, so you get
    a usage error.
    """

    result = runner.invoke(ape_cli, ["run", "error_forgot_click"])
    assert "Usage: cli run" in result.output


@skip_projects_except("with-contracts")
def test_uncaught_tx_err(ape_cli, runner, project):
    result = runner.invoke(ape_cli, ["run", "txerr"])
    assert '/scripts/txerr.py", line 12, in main' in result.output
    assert "contract.setNumber(5, sender=account)" in result.output
    assert "ERROR: (ContractLogicError) Transaction failed." in result.output


@skip_projects_except("script")
def test_scripts_module_already_installed(ape_cli, runner, project, mocker):
    """
    Make sure that if there is for some reason a python module names `scripts`
    installed, it does not interfere with Ape's scripting mechanism.
    """
    mock_scripts = mocker.MagicMock()
    mock_path = mocker.MagicMock()
    mock_path._path = "path/to/scripts"
    mock_scripts.__file__ = None
    mock_scripts.__path__ = mock_path
    sys.modules["scripts"] = mock_scripts

    result = runner.invoke(ape_cli, ["run"])
    assert result.exit_code == 0, result.output

    del sys.modules["scripts"]


@skip_projects_except("script")
def test_run_recompiles_if_needed(ape_cli, runner, project):
    """
    Ensure that when a change is made to a contract,
    when we run a script, it re-compiles the script first.
    """
    # Ensure we begin compiled.
    runner.invoke(ape_cli, ["compile", "--force"])

    # Make a change to the contract
    contract = project.contracts_folder / "contract.json"
    method_name = project.TestContractVy.contract_type.view_methods[0].name
    new_method_name = f"f__{method_name}__"
    new_contract_text = contract.read_text().replace(method_name, new_method_name)
    contract.write_text(new_contract_text)

    # Run the script. It better recompile first!
    result = runner.invoke(ape_cli, ["run", "output_contract_view_methods"])
    assert new_method_name in result.output



from typing import Any, Callable, TypeVar, cast

import pluggy

F = TypeVar("F", bound=Callable[..., Any])
hookimpl = cast(Callable[[F], F], pluggy.HookimplMarker("ape"))
hookspec = pluggy.HookspecMarker("ape")

plugin_manager = pluggy.PluginManager("ape")
"""A manager responsible for registering and accessing plugins (singleton)."""


class PluginType:
    """
    The base plugin class in ape. There are several types of plugins available in ape, such
    as the :class:`~ape.plugins.config.Config` or :class:`~ape.plugins.network.EcosystemPlugin`.
    Each one of them subclass this class. It is used to namespace the plugin hooks for the
    registration process, and to ensure overall conformance to type interfaces as much as possible.
    """



from pathlib import Path

import pytest

from ape.contracts import ContractContainer
from ape.exceptions import APINotImplementedError, CompilerError
from tests.conftest import skip_if_plugin_installed


def test_get_imports(project, compilers):
    # See ape-solidity for better tests
    assert not compilers.get_imports(project.source_paths)


def test_missing_compilers_without_source_files(project):
    result = project.extensions_with_missing_compilers()
    assert result == []


@skip_if_plugin_installed("vyper", "solidity")
def test_missing_compilers_with_source_files(project_with_source_files_contract):
    result = project_with_source_files_contract.extensions_with_missing_compilers()
    assert ".vy" in result
    assert ".sol" in result


@skip_if_plugin_installed("vyper", "solidity")
def test_missing_compilers_error_message(project_with_source_files_contract, sender):
    missing_exts = project_with_source_files_contract.extensions_with_missing_compilers()
    expected = (
        r"ProjectManager has no attribute or contract named 'ContractA'\. "
        r"However, there is a source file named 'ContractA', "
        r"did you mean to reference a contract name from this source file\? "
        r"Else, could it be from one of the missing compilers for extensions: "
        rf'{", ".join(sorted(missing_exts))}\?'
    )
    with pytest.raises(AttributeError, match=expected):
        project_with_source_files_contract.ContractA.deploy(
            sender.address, sender.address, sender=sender
        )


def test_get_compiler(compilers):
    compiler = compilers.get_compiler("ethpm")
    assert compiler.name == "ethpm"
    assert compilers.get_compiler("foobar") is None


def test_get_compiler_with_settings(compilers, mock_compiler, project_with_contract):
    existing_compilers = compilers._registered_compilers_cache[project_with_contract.path]
    all_compilers = {**existing_compilers, mock_compiler.ext: mock_compiler}

    try:
        compilers._registered_compilers_cache[project_with_contract.path] = all_compilers
        compiler_0 = compilers.get_compiler("mock", settings={"bar": "foo"})
        compiler_1 = compiler_0.get_compiler("mock", settings={"foo": "bar"})

    finally:
        compilers._registered_compilers_cache[project_with_contract.path] = existing_compilers

    assert compiler_0.compiler_settings != compiler_1.compiler_settings
    assert id(compiler_0) != id(compiler_1)


def test_getattr(compilers):
    compiler = compilers.ethpm
    assert compiler.name == "ethpm"

    with pytest.raises(AttributeError, match=r"No attribute or compiler named 'foobar'\."):
        _ = compilers.foobar


def test_supports_tracing(compilers):
    assert not compilers.ethpm.supports_source_tracing


def test_flatten_contract(compilers, project_with_contract):
    """
    Positive tests exist in compiler plugins that implement this behavior.b
    """
    source_id = project_with_contract.ApeContract0.contract_type.source_id
    path = project_with_contract.contracts_folder / source_id

    with pytest.raises(APINotImplementedError):
        compilers.flatten_contract(path)

    expected = r"Unable to flatten contract\. Missing compiler for '.foo'\."
    with pytest.raises(CompilerError, match=expected):
        compilers.flatten_contract(Path("contract.foo"))


def test_contract_type_collision(compilers, project_with_contract, mock_compiler):
    existing_compilers = compilers._registered_compilers_cache[project_with_contract.path]
    all_compilers = {**existing_compilers, mock_compiler.ext: mock_compiler}

    try:
        compilers._registered_compilers_cache[project_with_contract.path] = all_compilers

        # Make contracts of type .__mock__ with the same names.
        contract_name = next(iter(project_with_contract.contracts))
        new_contract = project_with_contract.path / f"{contract_name}{mock_compiler.ext}"
        new_contract.write_text("foobar")

        with pytest.raises(CompilerError, match="ContractType collision.*"):
            compilers.compile([new_contract])

    finally:
        compilers._registered_compilers_cache[project_with_contract.path] = existing_compilers


def test_compile_with_settings(mock_compiler, compilers, project_with_contract):
    existing_compilers = compilers._registered_compilers_cache[project_with_contract.path]
    all_compilers = {**existing_compilers, mock_compiler.ext: mock_compiler}
    new_contract = project_with_contract.path / f"AMockContract{mock_compiler.ext}"
    new_contract.write_text("foobar")
    settings = {"mock": {"foo": "bar"}}

    try:
        compilers._registered_compilers_cache[project_with_contract.path] = all_compilers
        compilers.compile([new_contract], settings=settings)

    finally:
        compilers._registered_compilers_cache[project_with_contract.path] = existing_compilers

    actual = mock_compiler.method_calls[0][2]["update"]["compiler_settings"]["mock"]
    assert actual == settings["mock"]


def test_compile_str_path(compilers, project_with_contract):
    path = next(iter(project_with_contract.source_paths))
    actual = compilers.compile([str(path)])
    contract_name = path.stem
    assert actual[contract_name].name == contract_name


def test_compile_source(compilers):
    code = '[{"name":"foo","type":"fallback", "stateMutability":"nonpayable"}]'
    actual = compilers.compile_source("ethpm", code)
    assert isinstance(actual, ContractContainer)



#!/usr/bin/env python
from pathlib import Path
from typing import Dict

from setuptools import find_packages, setup

here = Path(__file__).parent.absolute()
packages_data: Dict = {}
with open(here / "src" / "ape" / "__modules__.py", encoding="utf8") as modules_file:
    exec(modules_file.read(), packages_data)

extras_require = {
    "test": [  # `test` GitHub Action jobs uses this
        "pytest-xdist",  # multi-process runner
        "pytest-cov>=4.0.0,<5",  # Coverage analyzer plugin
        "pytest-mock",  # For creating mocks
        "hypothesis>=6.2.0,<7.0",  # Strategy-based fuzzer
        "hypothesis-jsonschema==0.19.0",  # JSON Schema fuzzer extension
    ],
    "lint": [
        "black>=23.11.0,<24",  # Auto-formatter and linter
        "mypy>=1.7.0,<2",  # Static type analyzer
        "types-PyYAML",  # Needed due to mypy typeshed
        "types-requests",  # Needed due to mypy typeshed
        "types-setuptools",  # Needed due to mypy typeshed
        "pandas-stubs==1.2.0.62",  # Needed due to mypy typeshed
        "types-SQLAlchemy>=1.4.49",  # Needed due to mypy typeshed
        "flake8>=6.1.0,<7",  # Style linter
        "flake8-breakpoint>=1.1.0,<2",  # detect breakpoints left in code
        "flake8-print>=4.0.1,<5",  # detect print statements left in code
        "isort>=5.10.1,<6",  # Import sorting linter
        "mdformat>=0.7.17",  # Auto-formatter for markdown
        "mdformat-gfm>=0.3.5",  # Needed for formatting GitHub-flavored markdown
        "mdformat-frontmatter>=0.4.1",  # Needed for frontmatters-style headers in issue templates
        "mdformat-pyproject>=0.0.1",  # Allows configuring in pyproject.toml
        "pydantic<2.0",  # Needed for successful type check. TODO: Remove after full v2 support.
    ],
    "doc": [
        "myst-parser>=1.0.0,<2",  # Parse markdown docs
        "sphinx-click>=4.4.0,<5",  # For documenting CLI
        "Sphinx>=6.1.3,<7",  # Documentation generator
        "sphinx_rtd_theme>=1.2.0,<2",  # Readthedocs.org theme
        "sphinxcontrib-napoleon>=0.7",  # Allow Google-style documentation
        "sphinx-plausible>=0.1.2,<0.2",
    ],
    "release": [  # `release` GitHub Action job uses this
        "setuptools",  # Installation tool
        "wheel",  # Packaging tool
        "twine==3.8.0",  # Package upload tool
    ],
    "dev": [
        # commitizen: Manage commits and publishing releases
        (here / "cz-requirement.txt").read_text().strip(),
        "pre-commit",  # Ensure that linters are run prior to committing
        "pytest-watch",  # `ptw` test watcher/runner
        "ipdb",  # Debugger (Must use `export PYTHONBREAKPOINT=ipdb.set_trace`)
    ],
    # NOTE: These are extras that someone can install to get up and running quickly w/ ape
    #       They should be kept up to date with what works and what doesn't out of the box
    #       Usage example: `pipx install eth-ape[recommended-plugins]`
    "recommended-plugins": (here / "recommended-plugins.txt").read_text().splitlines(),
}

# NOTE: `pip install -e .[dev]` to install package
extras_require["dev"] = (
    extras_require["test"]
    + extras_require["lint"]
    + extras_require["doc"]
    + extras_require["release"]
    + extras_require["dev"]
    # NOTE: Do *not* install `recommended-plugins` w/ dev
)

with open("./README.md") as readme:
    long_description = readme.read()


setup(
    name="eth-ape",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    description="Ape Ethereum Framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="ApeWorX Ltd.",
    author_email="admin@apeworx.io",
    url="https://apeworx.io",
    project_urls={
        "Documentation": "https://docs.apeworx.io/ape/",
        "Funding": "https://gitcoin.co/grants/5958/ape-maintenance-fund",
        "Source": "https://github.com/ApeWorX/ape",
        "Tracker": "https://github.com/ApeWorX/ape/issues",
        "Twitter": "https://twitter.com/ApeFramework",
    },
    include_package_data=True,
    install_requires=[
        "click>=8.1.6,<9",
        "ijson>=3.1.4,<4",
        "importlib-metadata",
        "ipython>=8.5.0,<9",
        "lazyasd>=0.1.4",
        "packaging>=23.0,<24",
        "pandas>=1.3.0,<2",
        "pluggy>=1.3,<2",
        "pydantic>=1.10.8,<3",
        "PyGithub>=1.59,<2",
        "pytest>=6.0,<8.0",
        "python-dateutil>=2.8.2,<3",
        "PyYAML>=5.0,<7",
        "requests>=2.28.1,<3",
        "rich>=12.5.1,<13",
        "SQLAlchemy>=1.4.35",
        "tqdm>=4.62.3,<5.0",
        "traitlets>=5.3.0",
        "urllib3>=1.26.16,<2",
        "watchdog>=3.0,<4",
        # ** Dependencies maintained by Ethereum Foundation **
        "eth-abi>=4.1.0,<5",
        "eth-account>=0.8,<0.9",
        "eth-typing>=3.4,<4",
        "eth-utils>=2.2.0,<3",
        "hexbytes>=0.2.3,<1",
        "py-geth>=3.13.0,<4",
        "web3[tester]>=6.7.0,<7",
        # ** Dependencies maintained by ApeWorX **
        "eip712>=0.2.1,<0.3",
        "ethpm-types>=0.5.10,<0.6",
        "evm-trace>=0.1.0a23",
    ],
    entry_points={
        "console_scripts": ["ape=ape._cli:cli"],
        "pytest11": ["ape_test=ape.pytest.plugin"],
        "ape_cli_subcommands": [
            "ape_accounts=ape_accounts._cli:cli",
            "ape_cache=ape_cache._cli:cli",
            "ape_compile=ape_compile._cli:cli",
            "ape_console=ape_console._cli:cli",
            "ape_plugins=ape_plugins._cli:cli",
            "ape_run=ape_run._cli:cli",
            "ape_networks=ape_networks._cli:cli",
            "ape_test=ape_test._cli:cli",
            "ape_init=ape_init._cli:cli",
            "ape_pm=ape_pm._cli:cli",
        ],
    },
    python_requires=">=3.8,<4",
    extras_require=extras_require,
    py_modules=packages_data["__modules__"],
    license="Apache-2.0",
    zip_safe=False,
    keywords="ethereum",
    packages=find_packages("src"),
    package_dir={"": "src"},
    package_data={p: ["py.typed"] for p in packages_data["__modules__"]},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
)



from typing import Iterator, Union

from eth_account import Account
from eth_account.messages import SignableMessage
from eth_utils import to_bytes, to_hex
from pydantic.dataclasses import dataclass

from ape.types import AddressType

# Fix 404 in doc link.
SignableMessage.__doc__ = (SignableMessage.__doc__ or "").replace(
    "EIP-191_", "`EIP-191 <https://eips.ethereum.org/EIPS/eip-191>`__"
)


def _left_pad_bytes(val: bytes, num_bytes: int) -> bytes:
    return b"\x00" * (num_bytes - len(val)) + val if len(val) < num_bytes else val


@dataclass(frozen=True)
class _Signature:
    v: int
    r: bytes
    s: bytes

    def __iter__(self) -> Iterator[Union[int, bytes]]:
        # NOTE: Allows tuple destructuring
        yield self.v
        yield self.r
        yield self.s

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} v={self.v} r={to_hex(self.r)} s={to_hex(self.s)}>"

    def encode_vrs(self) -> bytes:
        return to_bytes(self.v) + _left_pad_bytes(self.r, 32) + _left_pad_bytes(self.s, 32)

    def encode_rsv(self) -> bytes:
        return _left_pad_bytes(self.r, 32) + _left_pad_bytes(self.s, 32) + to_bytes(self.v)


class MessageSignature(_Signature):
    """
    A ECDSA signature (vrs) of a message.
    """


def recover_signer(msg: SignableMessage, sig: MessageSignature) -> AddressType:
    """
    Get the address of the signer.

    Args:
        :class:``SignableMessage``: A formatted and signable message.
        :class:`~ape.types.MessageSignature`MessageSignature: Signature of the message.

    Returns:
        ``AddressType``: address of message signer.
    """
    return Account.recover_message(msg, sig)


class TransactionSignature(_Signature):
    """
    A ECDSA signature (vrs) of a transaction.
    """


__all__ = ["MessageSignature", "TransactionSignature", "SignableMessage"]



from pathlib import Path
from typing import List

import pytest

from ape.types.coverage import (
    ContractCoverage,
    ContractSourceCoverage,
    CoverageProject,
    CoverageReport,
    CoverageStatement,
    FunctionCoverage,
)

STMT_0_HIT = 12
STMT_1_HIT = 9
STMT_2_HIT = 0


@pytest.fixture
def statements():
    return create_statements(20, 21, 21)


def create_statements(*pcs) -> List[CoverageStatement]:
    return [
        CoverageStatement(pcs={pcs[0]}, hit_count=STMT_0_HIT),
        CoverageStatement(pcs={pcs[1]}, hit_count=STMT_1_HIT),
        CoverageStatement(pcs={pcs[2]}, hit_count=STMT_2_HIT),
    ]


@pytest.fixture
def foo_function(statements):
    return FunctionCoverage(name="foo", full_name="foo()", statements=statements, hit_count=1)


@pytest.fixture
def bar_function():
    return FunctionCoverage(
        name="bar", full_name="bar()", statements=create_statements(30, 31, 32), hit_count=0
    )


@pytest.fixture
def function(foo_function):
    return foo_function  # alias for when there's only 1 function in test.


@pytest.fixture
def contract(foo_function, bar_function):
    return ContractCoverage(name="Contract", functions=[foo_function, bar_function])


@pytest.fixture
def second_contract(foo_function, bar_function):
    # It's ok to have the same functions.
    return ContractCoverage(name="Contract_Second", functions=[foo_function, bar_function])


@pytest.fixture
def source_contract(contract):
    return ContractSourceCoverage(source_id="Contract.vy", contracts=[contract])


@pytest.fixture
def second_source_contract(second_contract):
    return ContractSourceCoverage(source_id="Contract_Second.vy", contracts=[second_contract])


@pytest.fixture
def coverage_project(source_contract, second_source_contract):
    return CoverageProject(name="__local__", sources=[source_contract, second_source_contract])


@pytest.fixture
def coverage_report(coverage_project):
    return CoverageReport(source_folders=[Path.cwd()], projects=[coverage_project], timestamp=0)


class TestFunctionCoverage:
    def test_hit_count(self, function):
        assert function.hit_count == 1

    def test_lines_coverage(self, function):
        assert function.lines_covered == 2

    def test_miss_count(self, function):
        assert function.miss_count == 1  # stmt 3

    def test_line_rate(self, function):
        assert function.line_rate == 2 / 3

    def test_line_rate_when_no_statements(self):
        """
        An edge case: when a function has no statements, its line rate
        it either 0% if it was not called or 100% if it called.
        """
        function = FunctionCoverage(name="bar", full_name="bar()")
        assert function.hit_count == 0
        function.hit_count += 1
        assert function.line_rate == 1


class TestContractCoverage:
    def test_function_rate(self, contract):
        assert contract.function_rate == 0.5

    def test_lines_coverage(self, contract):
        assert contract.lines_covered == 4

    def test_miss_count(self, contract):
        assert contract.miss_count == 2

    def test_line_rate(self, contract):
        assert contract.line_rate == 2 / 3


class TestSourceCoverage:
    def test_function_rate(self, source_contract):
        assert source_contract.function_rate == 0.5

    def test_lines_coverage(self, source_contract):
        assert source_contract.lines_covered == 4

    def test_miss_count(self, source_contract):
        assert source_contract.miss_count == 2

    def test_line_rate(self, source_contract):
        assert source_contract.line_rate == 2 / 3


class TestCoverageProject:
    def test_function_rate(self, coverage_project):
        assert coverage_project.function_rate == 0.5

    def test_lines_coverage(self, coverage_project):
        # Doubles because has 2 sources in it now (with same amounts of things)
        assert coverage_project.lines_covered == 8

    def test_miss_count(self, coverage_project):
        assert coverage_project.miss_count == 4

    def test_line_rate(self, coverage_project):
        assert coverage_project.line_rate == 2 / 3


class TestCoverageReport:
    def test_function_rate(self, coverage_report):
        assert coverage_report.function_rate == 0.5

    def test_lines_coverage(self, coverage_report):
        assert coverage_report.lines_covered == 8

    def test_miss_count(self, coverage_report):
        assert coverage_report.miss_count == 4

    def test_line_rate(self, coverage_report):
        assert coverage_report.line_rate == 2 / 3



from ape_ethereum.transactions import Receipt, TransactionStatusEnum
from tests.conftest import explorer_test


def test_history(sender, receiver, chain):
    length_at_start = len(chain.history[sender].sessional)
    receipt = sender.transfer(receiver, "1 wei")
    transactions_from_cache = list(sender.history)
    assert len(transactions_from_cache) == length_at_start + 1
    assert sender.history[-1] == receipt
    assert sender.history[0:] == transactions_from_cache[0:]
    assert sender.history[:-1] == transactions_from_cache[:-1]

    txn = transactions_from_cache[-1]
    assert txn.sender == receipt.sender == sender
    assert txn.receiver == receipt.receiver == receiver


@explorer_test
def test_history_caches_sender_over_address_key(
    chain, eth_tester_provider, sender, vyper_contract_container, ethereum, mock_explorer
):
    # When getting receipts from the explorer for contracts, it includes transactions
    # made to the contract. This test shows we cache by sender and not address key.
    contract = sender.deploy(vyper_contract_container, 0)
    network = ethereum.local
    txn = ethereum.create_transaction(
        receiver=contract.address, sender=sender.address, value=10000000000000000000000
    )
    known_receipt = Receipt(
        block_number=10,
        gas_price=11,
        gas_used=12,
        gas_limit=13,
        status=TransactionStatusEnum.NO_ERROR.value,
        txn_hash="0x98d2aee8617897b5983314de1d6ff44d1f014b09575b47a88267971beac97b2b",
        transaction=txn,
    )

    # The receipt is already known and cached by the sender.
    chain.history.append(known_receipt)

    # We ask for receipts from the contract, but it returns ones sent to the contract.
    def get_txns_patch(address):
        if address == contract.address:
            yield from [known_receipt]

    mock_explorer.get_account_transactions.side_effect = get_txns_patch
    network.__dict__["explorer"] = mock_explorer
    eth_tester_provider.network = network

    # Previously, this would error because the receipt was cached with the wrong sender
    try:
        actual = [t for t in chain.history[contract.address].sessional]

        # Actual is 0 because the receipt was cached under the sender.
        assert len(actual) == 0
    finally:
        if "explorer" in network.__dict__:
            del network.__dict__["explorer"]



import pytest

from ape.api.address import BaseAddress
from ape.types import AddressType


@pytest.fixture
def custom_address(zero_address):
    class MyAddress(BaseAddress):
        @property
        def address(self) -> AddressType:
            return zero_address

    return MyAddress()


def test_eq(zero_address, custom_address):
    assert custom_address == zero_address
    assert zero_address == custom_address


def test_contains_in_list(zero_address, custom_address):
    assert zero_address in [custom_address]
    assert custom_address in [zero_address]



import pytest

from ape.exceptions import NetworkError
from ape.utils import DEFAULT_TEST_CHAIN_ID


class NewChainID:
    chain_id = 12345678987654

    def __call__(self, *args, **kwargs) -> int:
        self.chain_id += 1
        return self.chain_id


chain_id_factory = NewChainID()

DEFAULT_CHOICES = {
    "::geth",
    "::test",
    ":goerli",
    ":goerli:geth",
    ":sepolia",
    ":sepolia:geth",
    ":local",
    ":mainnet",
    ":mainnet:geth",
    "ethereum",
    "ethereum::test",
    "ethereum::geth",
    "ethereum:goerli",
    "ethereum:goerli:geth",
    "ethereum:sepolia",
    "ethereum:sepolia:geth",
    "ethereum:local",
    "ethereum:local:geth",
    "ethereum:local:test",
    "ethereum:mainnet",
    "ethereum:mainnet:geth",
}


@pytest.fixture
def get_provider_with_unused_chain_id(networks_connected_to_tester):
    networks = networks_connected_to_tester

    def fn(**more_settings):
        chain_id = chain_id_factory()
        settings = {"chain_id": chain_id, **more_settings}
        choice = "ethereum:local:test"
        disconnect_after = settings.pop("disconnect_after", True)
        context = networks.parse_network_choice(
            choice, disconnect_after=disconnect_after, provider_settings=settings
        )
        return context

    return fn


@pytest.fixture
def get_context(networks_connected_to_tester):
    def fn():
        return networks_connected_to_tester.parse_network_choice("ethereum:local:test")

    return fn


@pytest.fixture
def network_with_no_providers(ethereum):
    network = ethereum.get_network("goerli-fork")
    default_provider = network.default_provider
    providers = network.__dict__["providers"]

    if default_provider or providers:
        # Handle user running tests with forked-network plugins installed
        network._default_provider = None
        network.__dict__["providers"] = []

    yield network

    if default_provider or providers:
        network._default_provider = default_provider
        network.__dict__["providers"] = providers


def test_get_network_choices(networks, ethereum, mocker):
    # Simulate having a provider like foundry installed.
    mock_provider = mocker.MagicMock()
    ethereum.networks["mainnet-fork"].providers["mock"] = mock_provider
    ethereum.networks["local"].providers["mock"] = mock_provider

    # Ensure the provider shows up both mainnet fork and local
    # (There was once a bug where it was skipped!)
    expected = {":mainnet-fork:mock", ":local:mock"}

    actual = {c for c in networks.get_network_choices()}
    expected = DEFAULT_CHOICES.union(expected)
    assert expected.issubset(actual)


def test_get_network_choices_filter_ecosystem(networks):
    actual = {c for c in networks.get_network_choices(ecosystem_filter="ethereum")}
    assert DEFAULT_CHOICES.issubset(actual)


def test_get_network_choices_filter_network(networks):
    actual = {c for c in networks.get_network_choices(network_filter="mainnet")}
    mainnet_choices = {
        ":mainnet",
        ":mainnet:geth",
        "ethereum",
        "ethereum:mainnet",
        "ethereum:mainnet:geth",
    }
    assert mainnet_choices.issubset(actual)


def test_get_network_choices_filter_provider(networks):
    actual = {c for c in networks.get_network_choices(provider_filter="test")}
    expected = {"::test", ":local", "ethereum:local", "ethereum:local:test", "ethereum"}
    assert all(e in actual for e in expected)


def test_get_provider_when_no_default(network_with_no_providers):
    expected = f"No default provider for network '{network_with_no_providers.name}'"
    with pytest.raises(NetworkError, match=expected):
        # Not provider installed out-of-the-box for goerli-fork network
        provider = network_with_no_providers.get_provider()
        assert not provider, f"Provider should be None but got '{provider.name}'"


def test_repr_connected_to_local(networks_connected_to_tester):
    actual = repr(networks_connected_to_tester)
    expected = f"<NetworkManager active_provider=<test chain_id={DEFAULT_TEST_CHAIN_ID}>>"
    assert actual == expected

    # Check individual network
    actual = repr(networks_connected_to_tester.provider.network)
    expected = f"<ethereum:local chain_id={DEFAULT_TEST_CHAIN_ID}>"
    assert actual == expected


def test_repr_disconnected(networks_disconnected):
    assert repr(networks_disconnected) == "<NetworkManager>"
    assert repr(networks_disconnected.ethereum) == "<ethereum>"
    assert repr(networks_disconnected.ethereum.local) == "<ethereum:local>"
    assert repr(networks_disconnected.ethereum.goerli) == "<ethereum:goerli chain_id=5>"


def test_get_provider_from_choice_adhoc_provider(networks_connected_to_tester):
    uri = "https://geth:1234567890abcdef@geth.foo.bar/"
    provider = networks_connected_to_tester.get_provider_from_choice(f"ethereum:local:{uri}")
    assert provider.name == "geth"
    assert provider.uri == uri
    assert provider.network.name == "local"
    assert provider.network.ecosystem.name == "ethereum"


def test_get_provider_from_choice_adhoc_ecosystem(networks_connected_to_tester):
    uri = "https://geth:1234567890abcdef@geth.foo.bar/"
    provider = networks_connected_to_tester.get_provider_from_choice(uri)
    assert provider.name == "geth"
    assert provider.uri == uri
    assert provider.network.name == "adhoc"
    assert provider.network.ecosystem.name == "ethereum"


def test_parse_network_choice_same_provider(chain, networks_connected_to_tester, get_context):
    context = get_context()
    start_count = len(context.connected_providers)
    original_block_number = chain.blocks.height
    object_id = id(chain.provider)

    with context:
        assert id(chain.provider) == object_id
        count = len(context.connected_providers)

        # Does not create a new provider since it is the same chain ID
        assert count == start_count

    assert id(chain.provider) == object_id
    assert len(context.connected_providers) == start_count
    assert chain.blocks.height == original_block_number

    for provider in context.connected_providers.values():
        assert provider._web3 is not None


@pytest.mark.xdist_group(name="multiple-eth-testers")
def test_parse_network_choice_new_chain_id(get_provider_with_unused_chain_id, get_context):
    start_count = len(get_context().connected_providers)
    context = get_provider_with_unused_chain_id()
    with context:
        count = len(context.connected_providers)

        assert context._provider.chain_id != DEFAULT_TEST_CHAIN_ID

        # Creates new provider since it has a new chain ID
        assert count == start_count + 1

    for provider in context.connected_providers.values():
        assert provider._web3 is not None


@pytest.mark.xdist_group(name="multiple-eth-testers")
def test_disconnect_after(get_provider_with_unused_chain_id):
    context = get_provider_with_unused_chain_id()
    with context as provider:
        connection_id = provider.connection_id
        assert connection_id in context.connected_providers

    assert connection_id not in context.connected_providers


@pytest.mark.xdist_group(name="multiple-eth-testers")
def test_parse_network_choice_multiple_contexts(
    eth_tester_provider, get_provider_with_unused_chain_id
):
    first_context = get_provider_with_unused_chain_id()
    assert (
        eth_tester_provider.chain_id == DEFAULT_TEST_CHAIN_ID
    ), "Test setup failed - expecting to start on default chain ID"
    assert eth_tester_provider._make_request("eth_chainId") == DEFAULT_TEST_CHAIN_ID

    with first_context:
        start_count = len(first_context.connected_providers)
        expected_next_count = start_count + 1
        second_context = get_provider_with_unused_chain_id()
        with second_context:
            # Second context should already know about connected providers
            assert len(first_context.connected_providers) == expected_next_count
            assert len(second_context.connected_providers) == expected_next_count

    assert eth_tester_provider.chain_id == DEFAULT_TEST_CHAIN_ID
    assert eth_tester_provider._make_request("eth_chainId") == DEFAULT_TEST_CHAIN_ID


def test_getattr_ecosystem_with_hyphenated_name(networks, ethereum):
    networks.ecosystems["hyphen-in-name"] = networks.ecosystems["ethereum"]
    assert networks.hyphen_in_name  # Make sure does not raise AttributeError
    del networks.ecosystems["hyphen-in-name"]



from functools import singledispatchmethod
from typing import Iterator, Optional

from ape.api import ReceiptAPI
from ape.api.query import ContractCreationQuery, QueryAPI, QueryType
from ape.exceptions import QueryEngineError
from ape_geth.provider import BaseGethProvider


class OTSQueryEngine(QueryAPI):
    @singledispatchmethod
    def estimate_query(self, query: QueryType) -> Optional[int]:  # type: ignore[override]
        return None

    @singledispatchmethod
    def perform_query(self, query: QueryType) -> Iterator:  # type: ignore[override]
        raise QueryEngineError(
            f"{self.__class__.__name__} cannot handle {query.__class__.__name__} queries."
        )

    @estimate_query.register
    def estimate_contract_creation_query(self, query: ContractCreationQuery) -> Optional[int]:
        if provider := self.network_manager.active_provider:
            if not isinstance(provider, BaseGethProvider):
                return None
            elif uri := provider.http_uri:
                return 225 if uri.startswith("http://") else 600

        return None

    @perform_query.register
    def get_contract_creation_receipt(self, query: ContractCreationQuery) -> Iterator[ReceiptAPI]:
        if self.network_manager.active_provider and isinstance(self.provider, BaseGethProvider):
            if receipt := self.provider._get_contract_creation_receipt(query.contract):
                yield receipt



import os
import shutil
import tempfile
from pathlib import Path
from typing import Dict

import pytest

from ape.managers.project.dependency import NpmDependency


@pytest.fixture
def oz_dependencies_config():
    def _create_oz_dependency(version: str) -> Dict:
        return {
            "name": "OpenZeppelin",
            "version": version,
            "github": "OpenZeppelin/openzeppelin-contracts",
        }

    return {"dependencies": [_create_oz_dependency("3.1.0"), _create_oz_dependency("4.4.2")]}


@pytest.fixture
def local_dependency(project_with_dependency_config):
    return project_with_dependency_config.dependencies["testdependency"]["local"]


@pytest.fixture
def project_with_downloaded_dependencies(temp_config, config, oz_dependencies_config):
    manifests_directory = Path(__file__).parent / "data" / "manifests"
    oz_manifests = manifests_directory / "OpenZeppelin"
    oz_manifests_dest = config.packages_folder / "OpenZeppelin"

    if oz_manifests_dest.is_dir():
        shutil.rmtree(oz_manifests_dest)

    shutil.copytree(oz_manifests, oz_manifests_dest)
    with temp_config(oz_dependencies_config) as project:
        yield project


def test_two_dependencies_with_same_name(project_with_downloaded_dependencies):
    name = "OpenZeppelin"
    oz_310 = project_with_downloaded_dependencies.dependencies[name]["3.1.0"]
    oz_442 = project_with_downloaded_dependencies.dependencies[name]["4.4.2"]
    base_uri = "https://github.com/OpenZeppelin/openzeppelin-contracts/releases/tag"

    assert oz_310.version == "3.1.0"
    assert oz_310.name == name
    assert str(oz_310.uri) == f"{base_uri}/v3.1.0"
    assert oz_442.version == "4.4.2"
    assert oz_442.name == name
    assert str(oz_442.uri) == f"{base_uri}/v4.4.2"


def test_dependency_contracts_folder(config, local_dependency):
    """
    The local dependency fixture uses a longer contracts folder path.
    This test ensures that the contracts folder field is honored, specifically
    In the case when it contains sub-paths.
    """
    actual = local_dependency.contracts_folder
    assert actual == "source/v0.1"


def test_local_dependency(local_dependency, config):
    assert local_dependency.name == "testdependency"
    assert local_dependency.version_id == "local"
    assert str(local_dependency.uri).startswith("file://")


def test_access_dependency_contracts(project_with_downloaded_dependencies):
    name = "OpenZeppelin"
    oz_442 = project_with_downloaded_dependencies.dependencies[name]["4.4.2"]
    contract = oz_442.AccessControl
    assert contract.contract_type.name == "AccessControl"


@pytest.mark.parametrize("ref", ("main", "v1.0.0", "1.0.0"))
def test_dependency_using_reference(ref, recwarn, dependency_manager):
    dependency_config = {
        "github": "apeworx/testfoobartest",
        "name": "foobar",
        "ref": ref,
    }
    dependency = dependency_manager.decode_dependency(dependency_config)
    _ = dependency.cached_manifest
    assert dependency.version is None
    assert dependency.ref == ref
    assert str(dependency.uri) == f"https://github.com/apeworx/testfoobartest/tree/{ref}"

    # Tests against bug where we tried creating version objects out of branch names,
    # thus causing warnings to show in `ape test` runs.
    assert DeprecationWarning not in [w.category for w in recwarn.list]


def test_npm_dependency(mock_home_directory):
    name = "@gnosis.pm"
    package = "safe-singleton-factory"
    version = "1.0.0"
    dependency = NpmDependency(name=package, npm=f"{name}/{package}", version=version)
    with tempfile.TemporaryDirectory() as temp_dir:
        os.chdir(temp_dir)

        # Test with both local and global node modules install.
        for base in (Path(temp_dir), mock_home_directory):
            package_folder = base / "node_modules" / name / package
            contracts_folder = package_folder / "contracts"
            contracts_folder.mkdir(parents=True)
            package_json = package_folder / "package.json"
            package_json.write_text(f'{{"version": "{version}"}}')
            file = contracts_folder / "contract.json"
            source_content = '{"abi": []}'
            file.write_text(source_content)

            manifest = dependency.extract_manifest(use_cache=False)

            assert manifest.sources
            assert str(manifest.sources["contract.json"].content) == f"{source_content}\n"

            shutil.rmtree(package_folder)


def test_compile(project_with_downloaded_dependencies):
    name = "OpenZeppelin"
    oz_442 = project_with_downloaded_dependencies.dependencies[name]["4.4.2"]
    # NOTE: the test data is pre-compiled because the ape-solidity plugin is required.
    actual = oz_442.compile()
    assert len(actual.contract_types) > 0


def test_compile_with_extra_settings(dependency_manager, project):
    settings = {".json": {"evm_version": "paris"}}
    path = "__test_path__"
    base_path = project.path / path
    contracts_path = base_path / "contracts"
    contracts_path.mkdir(parents=True)
    (contracts_path / "contract.json").write_text('{"abi": []}')

    data = {"name": "FooBar", "local": path, "config_override": settings}
    dependency = dependency_manager.decode_dependency(data)
    assert dependency.config_override == settings
